import json

from langchain_core.messages import SystemMessage
from pydantic import BaseModel, field_validator

initial_prompt = """
# Context #
You are a helpful AI assistant that can call different tools to help you with yours
task.

# TOOLS #

Use default values when available.
These are the available tool you have access to:

{tools}

When calling the call-tool, you need to use the following syntax:
```
call-tool(
    server_name="vector_store",
    name="retrieve-document",
    arguments={{
        "query": "This is a sample query",
        "k": 5,
        "threshold": 0.5
    }}
)
```
Where server_name is the name of the server you want to call the tool from.
Where name is the name of the tool you want to call and arguments is the argument you
want to pass to the tool.

# RESOURCES #

These are the resources you have access to:

{resources}

When you need access to the resources, you can calling the read-resource-tool.
You will need to use the following syntax:
```
read-resource-tool(
    server_name="vector_store",
    name="list-resources",
    uri="postgres://public/schemas"
)
```
Where server_name is the name of the server you want to call the resource from.
Where name is the name of the resource you want to call.
Uri is the URI you want to fetch the resource from.
"""


class MspToolPrompt(BaseModel):
    """Data model for the MSP tool prompt."""

    tools: str
    resources: str

    @field_validator("tools", "resources", mode="before")
    @classmethod
    def parse_dict(cls, val: dict) -> str:
        """Parse the tools."""
        return json.dumps(val)

    def get_prompt(
        self,
    ) -> SystemMessage:
        """Format the prompt to return a system message."""
        system_prompt = initial_prompt.format(
            tools=self.tools, resources=self.resources
        )
        return SystemMessage(content=system_prompt)



import json
from asyncio import gather
from contextlib import asynccontextmanager
from logging import getLogger

from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from mcp.shared.exceptions import McpError
from mcp.types import Tool
from pydantic import BaseModel, Field

from .data_models.mcp_servers import McpServers

log = getLogger("mcp_services.get_servers")


class ConfigurationError(Exception):
    """Error raised when there is an issue with the configuration."""

    pass


class CallInput(BaseModel):
    """Data model for call tool input."""

    name: str = Field(
        ...,
        description="The name of the tool to call from the server.",
        examples=["retrieve-documents"],
    )
    arguments: dict = Field(
        ...,
        description="The arguments to pass to the tool. This should be inferred "
        "from the list tool schema.",
        examples=[{"query": "lloyd hamilton data engineer", "k": 1, "threshold": 0.5}],
    )


class LangChainMcp:
    """List all the available tools for all servers."""

    def __init__(self, config_path: str = "./claude_mcp_config.json"):
        self.config_path = config_path
        self._server_configs: McpServers | None = None

    @property
    def server_configs(self) -> McpServers:
        """Get the server configurations."""
        if self._server_configs is None:
            self._server_configs = self._load_config()
        return self._server_configs

    def _load_config(self) -> McpServers:
        """Load mcp server configurations from file."""
        if self._server_configs is None:
            try:
                with open(self.config_path) as f:
                    config_data = json.load(f)
                self._server_configs = McpServers(**config_data)
            except FileNotFoundError as e:
                raise ConfigurationError(
                    f"Configuration file not found: {self.config_path}"
                ) from e
            except json.JSONDecodeError as e:
                raise ConfigurationError(
                    f"Invalid JSON in configuration file: {self.config_path}"
                ) from e
        return self._server_configs

    @staticmethod
    @asynccontextmanager
    async def get_session(server_params: StdioServerParameters) -> ClientSession:
        """Get a client MCP session.

        Args:
            server_params (StdioServerParameters): The server parameters.

        Yields:
            ClientSession: The client session for the MCP server.

        """
        async with (
            stdio_client(server_params) as (read, write),
            ClientSession(read, write) as session,
        ):
            await session.initialize()
            yield session

    async def _fetch_server_tools(
        self, server_name: str, params: StdioServerParameters
    ) -> tuple[str, Tool]:
        """List the available tools to call.

        Args:
            server_name (Server): The name of the server to create a session.
            params (StdioServerParameters): The server parameters.

        Returns:
            tuple[str, Tool]: The server name and the available tools.

        """
        log.info(f"Listing tools for server: {server_name}")
        async with self.get_session(params) as session:
            list_tool_results = await session.list_tools()
            return server_name, list_tool_results

    async def fetch_all_server_tools(self) -> dict:
        """List the available tools for all servers.

        Returns:
            dict: The available tools for all servers in the format
                {server_name: tools}.
        """
        coroutines = [
            self._fetch_server_tools(server_name, params)
            for server_name, params in self.server_configs.mcpServers.items()
        ]
        results = await gather(*coroutines)
        return {name: tools.model_dump(mode="json") for name, tools in results}

    async def _list_server_resources(
        self, server_name: str, params: StdioServerParameters
    ) -> tuple[str, Tool]:
        """List the available resources on a server.

        Args:
            server_name (Server): The name of the server to create a session.
            params (StdioServerParameters): The server parameters.

        Returns:
            tuple[str, Tool]: The server name and the available resources.

        """
        log.info(f"Listing resources for server: {server_name}")
        async with self.get_session(params) as session:
            try:
                list_resources_results = await session.list_resources()
            except McpError as e:
                log.warning(f"Error listing resources for server {server_name}: {e}")
                list_resources_results = []
            return server_name, list_resources_results

    async def list_all_server_resources(self) -> dict:
        """List the available resources for all servers.

        Return a dictionary of server names and their available resources. If
        there is an error listing resources for a server, the server name will
        be logged and the resources will not be included in the output.

        Returns:
            dict: The available resources for all servers in the format
                {server_name: resources}.
        """
        coroutines = [
            self._list_server_resources(server_name, params)
            for server_name, params in self.server_configs.mcpServers.items()
        ]
        results = await gather(*coroutines)
        resource_dict = {}
        for name, resources in results:
            if not resources:
                log.warning(f"No resources found for server: {name}")
                continue
            resource_dict[name] = resources.model_dump(mode="json")
        return resource_dict



import logging

from langchain_core.tools import tool
from mcp.types import AnyUrl
from pydantic import BaseModel, Field

from langchain_mcp_connect.get_servers import LangChainMcp

log = logging.getLogger("mcp_call_tool")


class CallInput(BaseModel):
    """Data model for call tool input."""

    server_name: str = Field(
        ...,
        description="The name of the server.",
        examples=["vector_store"],
    )
    name: str = Field(
        ...,
        description="The name of the tool to call from the server.",
        examples=["retrieve-documents"],
    )
    arguments: dict = Field(
        ...,
        description="The arguments to pass to the tool. This should be inferred "
                    "from the list tool schema.",
        examples=[{"query": "lloyd hamilton data engineer", "k": 1, "threshold": 0.5}],
    )


class ReadResourceInput(BaseModel):
    """Data model for list resource input."""

    server_name: str = Field(
        ...,
        description="The name of the server.",
        examples=["vector_store"],
    )
    name: str = Field(
        ...,
        description="The name of the tool to call from the server.",
        examples=["retrieve-documents"],
    )
    uri: str = Field(
        ...,
        description="The URI to list resources from.",
        examples=["postgres://public/schemas"],
    )


def load_mcp_params() -> LangChainMcp:
    """Load the server parameters."""
    return LangChainMcp()


@tool("call-tool", args_schema=CallInput)
async def call_tool(server_name: str, name: str, arguments: dict) -> list:
    """Call a tool with the given name and arguments.

    Args:
        server_name: The name of the server.
        name: The name of the tool to call.
        arguments: The arguments to pass to the tool.

    Example:
        call_tool(
            "vector-store-server",
            "retrieve-documents",
            {"query": "lloyd hamilton data engineer", "k": 1, "threshold": 0.5}
        )

    Returns:
        list: The list of results from the tool.
    """
    log.info(
        f"Calling tool {name} with arguments {arguments} from server {server_name}"
    )
    mcp = load_mcp_params()
    params = mcp.server_configs.mcpServers.get(server_name)
    async with mcp.get_session(params) as session:
        return await session.call_tool(name, arguments)


@tool("read-resources-tool", args_schema=ReadResourceInput)
async def read_resources_tool(server_name: str, name: str, uri: str) -> list:
    """List the resources available on a server.

    Args:
        server_name: The name of the server.
        name: The name of the tool to call.
        uri: The URI to list resources from.

    Example:
        read_resources_tool(
            server_name="postgres",
            name="table_names",
            uri="postgres://public/schemas"
        )
    """
    log.info(f"Reading resource {name} from server {server_name} at {uri}")
    mcp = load_mcp_params()
    params = mcp.server_configs.mcpServers.get(server_name)
    async with mcp.get_session(params) as session:
        return await session.read_resource(AnyUrl(uri))
